import argparse
import platform
import zipfile
import re
import sys
import shutil
from pathlib import Path

import reCBZ
import reCBZ.config as config
from reCBZ import wrappers, util
from reCBZ.profiles import ProfileDict


def print_title() -> None:
    align = int(config.term_width() / 2) - 11
    if align > 21: align = 21
    if align + 22 > config.term_width() or align < 0:
        align = 0
    align = align * ' '
    title_multiline = (f"{align}┬─┐┌─┐┌─┐┌┐ ┌─┐ ┌─┐┬ ┬ ──┐\n"
                       f"{align}├┬┘├┤ │  ├┴┐┌─┘ ├─┘└┬┘┌──┘\n"
                       f"{align}┴└─└─┘└─┘└─┘└─┘o┴   ┴ └──")
    print(title_multiline)


def unix_like_glob(arglist:list) -> list:
    """What we're essentially trying to supplant here is powershell's obtuse
    Get-ChildItem 'pattern' | foreach {cmd $_.FullName}, because it's just
    ridiculous expecting users to memorize that when a simple asterisk could do
    the job. PS if you're using Windows this is a reminder your OS SUCKS ASS"""
    import glob

    new = []
    for arg in arglist:
        if '*' in arg:
            new.extend(glob.glob(str(arg)))
        else:
            new.append(arg)

    return new


def main():
    # o god who art in heaven please guard mine anime girls
    mutually_exclusive_groups = []
    wiki='https://github.com/avalonv/reCBZ/wiki'
    desc='''accepted formats: .zip, .epub, .cbz. repacks to .cbz by default'''
    parser = argparse.ArgumentParser(
            prog=reCBZ.CMDNAME,
            usage="%(prog)s [options] files",
            description=desc,
            epilog=f"for detailed documentation, see {wiki}",)

    parser.add_argument( "-v", "--verbose", # log_group
        default=None,
        dest="loglevel",
        action="count",
        help="increase verbosity of messages, repeatable: -vv")
    parser.add_argument( "-s", "--silent", # log_group
        default=None,
        const=-1,
        dest="loglevel",
        action="store_const",
        help="disable progress messages")
    parser.add_argument( "-d", "--dry",
        default=None,
        dest="no_write",
        action="store_true",
        help="dry run, no changes are saved at the end (safe)")
    parser.add_argument( "-O", "--overwrite",
        default=None,
        dest="overwrite",
        action="store_true",
        help="overwrite the original archive")
    parser.add_argument( "-F", "--force",
        default=None,
        dest="force_write",
        action="store_true",
        help="write archive even if there are page errors (dangerous)")
    # parser.add_argument( "-u", "--unpack", # mode_group
    #     default=None,
    #     const='unpack',
    #     dest="mode",
    #     action="store_const",
    #     help="unpack the archive to the current directory (safe)")
    parser.add_argument( "--compare", # mode_group
        default=None,
        const='compare',
        dest="mode",
        action="store_const",
        help=argparse.SUPPRESS)
    parser.add_argument( "-a", "--assist", # mode_group
        default=None,
        const='assist',
        dest="mode",
        action="store_const",
        help="calculate size for each image format, ask which to use")
    parser.add_argument( "-A" ,"--auto", # mode_group
        default=None,
        const='auto',
        dest="mode",
        action="store_const",
        help="calculate size for each image format, pick smallest one")
    parser.add_argument( "-J" ,"--join", # mode_group
        default=None,
        const='join',
        dest="mode",
        action="store_const",
        help="append the contents of each file to the leftmost file")
    parser.add_argument( "--noprev",
        default=False,
        dest="noprev",
        action="store_true",
        help="ignore previously repacked files")
    log_group = ('verbose', 'silent')
    mutually_exclusive_groups.append(log_group)
    mode_group = ('compare', 'assist', 'a', 'auto', 'A', 'join', 'J')
    mutually_exclusive_groups.append(mode_group)

    archive_group = parser.add_argument_group(title="archive options")
    archive_group.add_argument( "-p", "--profile",
        default=None,
        # best to handle it externally as we always convert to upper case
        # choices=([prof.nickname for prof in profiles_list]),
        metavar="",
        dest="profile",
        type=str,
        help="target eReader profile. run --profiles to see options")
    archive_group.add_argument( "--profiles",
        default=False,
        dest="show_profiles",
        action='store_true',
        help=argparse.SUPPRESS)
    archive_group.add_argument( "--epub", # ext_group
        default=None,
        const='epub',
        dest="archive_format",
        action="store_const",
        help="save archive as epub")
    archive_group.add_argument( "--zip", # ext_group
        default=None,
        const='zip',
        dest="archive_format",
        action="store_const",
        help="save archive as zip")
    archive_group.add_argument( "--cbz", # ext_group
        default=None,
        const='cbz',
        dest="archive_format",
        action="store_const",
        help="save archive as cbz")
    archive_group.add_argument( "--compress",
        default=None,
        dest="compress_zip",
        action="store_true",
        help="attempt to further compress the archive when repacking")
    archive_group.add_argument( "--rtl",
        default=None,
        dest="right_to_left",
        action="store_true",
        help="sort pages from right to left. only affects epub")
    ext_group = ('epub', 'zip', 'cbz')
    mutually_exclusive_groups.append(ext_group)

    images_group = parser.add_argument_group(title="image options")
    images_group.add_argument( "-c", "--convert",
        default=None,
        choices=('jpeg', 'png', 'webp', 'webpll'),
        metavar="",
        dest="img_format",
        type=str,
        help="format to convert pages to: jpeg, webp, webpll, or png")
    images_group.add_argument( "--imgfmt", # deprecated
        default=None,
        choices=('jpeg', 'png', 'webp', 'webpll'),
        metavar="",
        dest="img_format",
        type=str,
        help=argparse.SUPPRESS)
    images_group.add_argument( "--quality",
        default=None,
        choices=(range(1,101)),
        metavar="0-95",
        dest="img_quality",
        type=int,
        help="save quality for lossy formats. >90 not recommended")
    images_group.add_argument( "--size",
        default=None,
        metavar="WidthxHeight",
        dest="size_str",
        type=str,
        help="rescale images to the specified resolution")
    images_group.add_argument( "--noup", # rescale_group
        default=None,
        dest="no_upscale",
        action="store_true",
        help="disable upscaling with --size")
    images_group.add_argument( "--nodown", # rescale_group
        default=None,
        dest="no_downscale",
        action="store_true",
        help="disable downscaling with --size")
    images_group.add_argument( "--nowebp",
        default=None,
        const=f'{config.blacklisted_fmts} webp webpll',
        dest="blacklisted_fmts",
        action="store_const",
        help="exclude webp from --auto and --assist")
    images_group.add_argument( "--bw", # color_group
        default=None,
        dest="grayscale",
        action="store_true",
        help="convert images to grayscale")
    images_group.add_argument( "--color", # color_group
        default=None,
        dest="grayscale",
        action="store_false",
        help="preserve color when using --profile")
    rescale_group = ('noup', 'nodown')
    color_group = ('bw', 'color')
    mutually_exclusive_groups.append(rescale_group)
    mutually_exclusive_groups.append(color_group)

    others_group = parser.add_argument_group(title="other")
    others_group.add_argument("--process", # process_group
        default=None,
        choices=(range(1,65)),
        metavar="2-64",
        dest="processes",
        type=int,
        help="maximum number of processes to spawn")
    others_group.add_argument( "--sequential", # process_group
        default=None,
        const=1,
        dest="processes",
        action="store_const",
        help="disable multiprocessing")
    others_group.add_argument( "--config",
        default=None,
        dest="show_config",
        action="store_true",
        help="show settings and exit")
    others_group.add_argument( "--version",
        default=None,
        dest="show_version",
        action="store_true",
        help="show version and exit")
    process_group = ('process', 'sequential')
    mutually_exclusive_groups.append(process_group)

    args, unknown_args = parser.parse_known_args()

    if args.show_version:
        print(f'{reCBZ.CMDNAME} v{reCBZ.__version__}')
        exit(0)

    # this is admittedly rather dumb, but we're trying to overcome argparse's
    # inability to include options in both groups and mutually_exclusive_groups
    # at the same time
    # more details here: https://github.com/python/cpython/issues/55193
    bad_args = []
    for mutex_group in mutually_exclusive_groups:
        group_matches = []
        for arg in sys.argv:
            for mutex_arg in mutex_group:
                p = re.compile(f'^[-]{{1,2}}{mutex_arg}')
                if p.match(arg) and arg not in group_matches:
                    group_matches.append(arg)
        if len(group_matches) >= 2: # only handle one pair at a time
            bad_args.append(group_matches[:2])

    if len(bad_args) >= 1:
        arg1, arg2 = bad_args[0] # only handle one group at a time
        print(f'{reCBZ.CMDNAME}: error: argument {arg1} not allowed with argument {arg2}')
        exit(1)

    # set profile first, ensure it can be overridden by explicit options
    if args.profile is not None:
        prof_name = args.profile.upper()
        try:
            config.set_profile(prof_name)
        except ValueError:
            print(f'{reCBZ.CMDNAME}: profile: invalid option "{prof_name}"')
            exit(1)

    if args.size_str is not None:
        newsize = args.size_str.lower().strip()
        try:
            newsize = tuple(map(int,newsize.split('x')))
            assert len(newsize) == 2
            config.img_size = newsize
        except (ValueError, AssertionError):
            print(f'{reCBZ.CMDNAME}: size: invalid option "{args.size_str}"')
            exit(1)

    # this is probably not the most pythonic way to do this
    # I'm sorry guido-san...
    for key, val in args.__dict__.items():
        if key in config.__dict__.keys() and val is not None:
            setattr(config, key, val)

    if args.show_config:
        for section in config._cfg.items():
            print(f'\n{section[0].upper()}:')
            for key, val in section[1].items():
                modified = config.__dict__[key]
                print(f"{key} =".ljust(18),
                      f"'{modified}'".ljust(8),
                      f"(default '{val}')")
        defaults_path = Path.joinpath(reCBZ.MODULE_PATH, 'defaults.toml')
        print(f'\nConfig location: {defaults_path}')
        exit(0)

    if args.show_profiles:
        print(f'{reCBZ.CMDNAME} -p ...')
        for prof_key, prof_class in ProfileDict.items():
            print(prof_key, '=', prof_class.desc)
        exit(0)

    # parse files
    if platform.system() == 'Windows':
        unknown_args = unix_like_glob(unknown_args)
    paths = []
    for arg in unknown_args:
        if Path(arg).is_file():
            paths.append(arg)
        elif Path(arg).is_dir():
            print(f'{reCBZ.CMDNAME}: {arg}: is a directory')
            parser.print_usage()
            exit(1)
        else:
            parser.print_help()
            print(f'\nunknown file or option: {arg}')
            exit(1)

    if len(paths) <= 0:
        print(f'{reCBZ.CMDNAME}: missing input file (see --help)')
        parser.print_usage()
        exit(1)

    if args.mode == 'join':
        if not len(paths) >= 2:
            print(f'{reCBZ.CMDNAME}: join: at least two files are needed')
            exit(1)

    if args.noprev:
        new = []
        for filename in paths:
            comment = zipfile.ZipFile(filename).comment
            if not comment == str.encode(config.ZIPCOMMENT):
                new.append(filename)
        diff = len(paths) - len(new)
        if diff > 0:
            print(f'{reCBZ.CMDNAME}: noprev: ignoring {diff} files')
            if len(new) == 0:
                exit(1)
            else:
                paths = new

    # everything passed. do stuff
    exit_code = 0
    if reCBZ.SHOWTITLE and config.loglevel >= 0: print_title()
    try:
        if args.mode == 'join':
            wrappers.join_archives(paths[0], paths[1:])
        for filename in paths:
            try:
                if args.mode is None:
                    wrappers.repack_archive(filename)
                elif args.mode == 'unpack':
                    wrappers.unpack_archive(filename)
                elif args.mode == 'compare':
                    wrappers.compare_fmts_archive(filename)
                elif args.mode == 'assist':
                    wrappers.assist_repack_archive(filename)
                elif args.mode == 'auto':
                    wrappers.auto_repack_archive(filename)
            except (wrappers.AbortedRepackError, wrappers.AbortedCompareError):
                exit_code = 2
                continue
    except (KeyboardInterrupt, util.MPrunnerInterrupt):
        print('\nGoooooooooodbye')
        exit(1)
    except wrappers.AbortedRepackError:
        exit_code = 2
    finally:
        g_cache = reCBZ.GLOBAL_CACHEDIR
        if g_cache.exists():
            try:
                util.mylog(f'cleanup(): {g_cache}')
                shutil.rmtree(g_cache)
            except PermissionError:
                util.mylog(f"PermissionError, couldn't clean {g_cache}")

    return exit_code

if __name__ == '__main__':
    # catching errors here won't always work, presumably because of namespace
    # mangling
    exit(main())



# Copyright 2022 Aleksandar Erkalović, Gwyn Vales
# This program is free software: you can redistribute it and/or modify it under
# the terms of the Affero GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License for more
# details.
# You should have received a copy of the Affero GNU General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
#
# 'ebooklib', authored by Aleksandar Erkalović, is licensed under the AGPL, and we
# are linking to it, therefore AGPL applies to this module as well. You may NOT
# distirubte derivatives using this module (i.e. epub.py) without also
# distributing the source code, even over a network, and any modifications to
# either epub.py or ebooklib must be licensed under the AGPL as well.
# non-commerical private use is explicitly permitted.
# WORK IN PROGRESS
from uuid import uuid4

import reCBZ.config as config
from ebooklib import epub
from reCBZ.util import mylog

POP_COVER = True


def single_chapter_epub(name: str, pages: list) -> str:
    book = epub.EpubBook()

    # attempt to distinguish author / title
    if ' - ' in name:
        title, author = name.split(' - ', 1)
    else:
        title = name
        author = 'reCBZ'
    book.set_title(title)
    book.add_author(author)
    book.set_language('en')
    book.set_identifier(str(uuid4()))

    # repacking the same file many times over would lead to additional copies
    # if we did include it
    if POP_COVER:
        cover = pages.pop(0)
    else:
        cover = pages[0]
    covert_ops = f'cover{cover.fmt.ext[0]}'
    with open(cover.fp, 'rb') as cover_file:
        book.set_cover(covert_ops, cover_file.read())

    spine = []
    for page_i, page in enumerate(pages, start=1):
        static_dest = f'static/{page_i}{page.fmt.ext[0]}'
        mime_type = page.fmt.mime
        if config.ebook_profile is not None:
            if page.landscape:
                height, width = config.ebook_profile.size
            else:
                width, height = config.ebook_profile.size
        else:
            width, height = page.size
        size_str = f'width={width} height={height}'
        mylog(f'writing {page.fp} to {static_dest} as {mime_type}')

        item = epub.EpubHtml(title=f'Page {page_i}',
                             file_name=f'page_{page_i}.xhtml', lang='en')
        item.content = f'''<html>
                            <head></head>
                            <body>
                                <img src="{static_dest}" {size_str}'/>
                            </body>
                         </html>'''

        with open(page.fp, 'rb') as image_file:
            image_content = image_file.read()
        # store read content relative to zip
        static_img = epub.EpubImage(uid=f'image_{page_i}', file_name=static_dest,
                                    media_type=mime_type, content=image_content)
        book.add_item(item)
        book.add_item(static_img)
        spine.append(item)
    book.toc.append(spine[0])

    # add navigation files
    book.add_item(epub.EpubNcx())
    book.add_item(epub.EpubNav())
    book.spine = (page for page in spine)

    if config.right_to_left is True:
        book.set_direction('rtl')

    if config.ebook_profile is not None:
        for tag in config.ebook_profile.epub_properties:
            book.add_metadata(*tag)
        source_fp = f'{name}{config.ebook_profile.epub_ext}'
    else:
        source_fp = f'{name}.epub'

    epub.write_epub(source_fp, book, {})
    return source_fp


from PIL import Image


class LossyFmt():
    lossless:bool = False
    quality:int = 80


class LosslessFmt():
    lossless:bool = True
    quality:int = 100


class Jpeg(LossyFmt):
    name:str = 'jpeg'
    ext:tuple = '.jpeg', '.jpg'
    desc:str = 'JPEG'
    mime:str = 'image/jpeg'

    @classmethod
    def save(cls, img:Image.Image, dest):
        img.save(dest, format='JPEG', optimize=True, quality=cls.quality)


class WebpLossy(LossyFmt):
    # conclusions: optmize appears to have no effect. method >4 has a very mild
    # effect (~1% reduction with 800MB jpeg source), but takes twice as long
    name:str = 'webp'
    ext:tuple = '.webp',
    desc:str = 'WebP'
    mime:str = 'image/webp'

    @classmethod
    def save(cls, img:Image.Image, dest):
        img.save(dest, fomrat='WEBP', lossless=cls.lossless, method=5, quality=cls.quality)


class WebpLossless(LosslessFmt):
    name:str = 'webpll'
    ext:tuple = '.webp',
    desc:str = 'WebP Lossless'
    mime:str = 'image/webp'

    @classmethod
    def save(cls, img:Image.Image, dest):
        # for some reason 'quality' is akin to Png compress_level when lossless
        img.save(dest, format='WEBP', lossless=cls.lossless, method=4, quality=100)


class Png(LosslessFmt):
    name:str = 'png'
    ext:tuple = '.png',
    desc:str = 'PNG'
    mime:str = 'image/png'

    @classmethod
    def save(cls, img:Image.Image, dest):
        img.save(dest, format='PNG', optimize=True, compress_level=9)


FormatList = (Jpeg, WebpLossy, WebpLossless, Png)
FormatDict = {cls.name:cls for cls in FormatList}



import re
import time
from pathlib import Path

import reCBZ
import reCBZ.config as config
from PIL import UnidentifiedImageError
from reCBZ.archive import ComicArchive
from reCBZ.util import human_bytes, pct_change, shorten, mylog

actual_stem = ''


class AbortedRepackError(IOError):
    """Some files couldn't be converted and are missing from the archive"""


class AbortedCompareError(IOError):
    """Caught PIL.UnidentifiedImageError in Archive.compute_fmt_sizes"""


def pprint_fmt_stats(base:tuple, totals:tuple) -> None:
    lines = f'┌─ Disk size ({config.samples_count}' + \
             ' pages) with present settings:\n'
    # justify to the left and right respectively. effectively the same
    # as using f'{part1: <25} | {part2: >8}\n'
    part1 = f'│   {base[1]}'.ljust(37)
    part2 = f'{human_bytes(base[0])}'.rjust(8)
    lines += f'{part1} {part2} |  0.00%\n'
    for i, total in enumerate(totals):
        if i == len(totals)-1:
            prefix = '└─'
        else:
            prefix = '├─'
        change = pct_change(base[0], total[0])
        part1 = f'{prefix}{i+1} {total[1]}'.ljust(37)
        part2 = f'{human_bytes(total[0])}'.rjust(8)
        lines += f'{part1} {part2} | {change}\n'
    mylog('', progress=True)
    print(lines[0:-1]) # strip last newline


def pprint_repack_stats(source:dict, new:dict, start_t:float) -> None:
    end_t = time.perf_counter()
    elapsed = f'{end_t - start_t:.2f}s'
    max_width = config.term_width()
    new_size = new['size']
    source_size = source['size']
    name = new['name']
    op = f"{source['type'].upper()} -> {new['type'].upper()}"
    if len(name) > max_width - 36:
        name = shorten(name, width=max_width - 36)
    if new_size > source_size:
        verb = 'INCREASE!'
    else:
        verb = 'decrease'
    change = pct_change(source_size, new_size)
    line1 = f"┌─ {op}: '{name}' completed in {elapsed}\n"
    line2 = f"└───■■ Source: {human_bytes(source_size)} ■ New: " +\
            f"{human_bytes(new_size)} ■ {change} {verb} ■■"
    length = len(line1) if len(line1) > len(line2) else len(line2)
    splitter = ''.rjust(length, '-')
    lines = line1 + line2 + '\n' + splitter
    mylog('', progress=True)
    if config.loglevel >= 0: print(lines)


def save(book):
    global actual_stem
    bad_files = book.bad_files
    if len(bad_files) > 0:
        if re.compile('\\.epub$').match(book.fp.suffix):
            old_len = len(bad_files)
            bad_files = [f for f in bad_files if not reCBZ.EPUB_FILES.match(str(f))]
            filtered = old_len - len(bad_files)
            if config.loglevel >= 0:
                print(f'[i] EPUB: filtered {filtered} files')
        if not config.force_write and len(bad_files) > 0:
            print(f'{book.fp.name}:')
            [print(f'error: {file.name}') for file in bad_files]
            print(f"[!] {len(bad_files)} files couldn't be converted")
            print('[!] Aborting (--force not specified)')
            print(''.rjust(config.term_width(), '^'))
            raise AbortedRepackError

    # fix suffix when source has two suffixes (kobo epub)
    try:
        actual_stem = reCBZ.KEPUB_EPUB.match(book.fp.name).group(0)
        # suffix = '.kepub.epub'
    except AttributeError:
        actual_stem = book.fp.stem
        # suffix = book.fp.suffix
    if not config.no_write:
        if config.overwrite:
            name = str(Path.joinpath(book.fp.parents[0], actual_stem))
            book.fp.unlink()
        # elif savedir TODO
        else:
            name = str(Path.joinpath(Path.cwd(), f'{actual_stem} [reCBZ]'))
        new_fp = Path(book.write_archive(config.archive_format, file_name=name))
    else:
        new_fp = book.fp
    book.cleanup()
    return str(new_fp)


def compare_fmts_archive(fp:str, quiet=False) -> tuple:
    """Run a sample with each image format, return the results"""
    try:
        results = ComicArchive(fp).compute_fmt_sizes()
    except UnidentifiedImageError as err:
        print("[!] Can't calculate size: PIL.UnidentifiedImageError. Aborting")
        raise AbortedCompareError
    if not quiet:
        pprint_fmt_stats(results[0], results[1:])
    return results


def unpack_archive(fp:str) -> None:
    # not implemented yet
    """Unpack the archive, converting all images within
    Returns path to repacked archive"""
    if config.loglevel >= 0: print(shorten('[i] Unpacking', fp))
    unpacked = ComicArchive(fp).extract()
    for file in unpacked:
        print(file)
    exit(1)


def repack_archive(fp:str) -> str:
    """Repack the archive, converting all images within
    Returns path to repacked archive"""
    if config.loglevel >= 0: print(shorten('[i] Repacking', fp))
    source_fp = Path(fp)
    start_t = time.perf_counter()
    book = ComicArchive(str(source_fp))
    book.extract()
    source_stats = {'name':source_fp.stem,
                    'size':source_fp.stat().st_size,
                    'type':source_fp.suffix[1:]}
    book.convert_pages() # page attributes are inherited from Config at init
    new_fp = Path(save(book))
    new_stats = {'name':new_fp.name,
                 'size':new_fp.stat().st_size,
                 'type':new_fp.suffix[1:]}
    pprint_repack_stats(source_stats, new_stats, start_t)
    return str(new_fp)


def join_archives(main_path:str, paths:list) -> str:
    """Concatenates the contents of paths to main_path and repacks
    Returns path to concatenated archive"""
    if config.loglevel >= 0: print(shorten('[i] Repacking', main_path))
    source_fp = Path(main_path)
    start_t = time.perf_counter()
    main_book = ComicArchive(main_path)
    sum_size = sum(Path(file).stat().st_size for file in paths)
    source_stats = {'name':source_fp.stem,
                    'size':sum_size,
                    'type':source_fp.suffix[1:]}
    for file in paths:
        book = ComicArchive(file)
        main_book.add_chapter(book)
    main_book.convert_pages()
    new_fp = Path(save(main_book))
    new_stats = {'name':new_fp.name,
                 'size':new_fp.stat().st_size,
                 'type':new_fp.suffix[1:]}
    pprint_repack_stats(source_stats, new_stats, start_t)
    main_book.cleanup()
    return str(new_fp)


def assist_repack_archive(fp:str) -> str:
    """Run a sample with each image format, then ask which to repack
    the rest of the archive with
    Returns path to repacked archive"""
    results = compare_fmts_archive(fp)
    options_dic = {i : total[2] for i, total in enumerate(results[1:])}
    metavar = f'[1-{len(options_dic)}]'
    while True:
        try:
            reply = int(input(f"■─■ Proceed with {metavar}: ")) - 1
            selection = options_dic[reply]
            break
        except (ValueError, KeyError):
            print('[!] Ctrl+C to cancel')
            continue
        except KeyboardInterrupt:
            print('[!] Aborting')
            exit(1)
    config.img_format = selection
    return repack_archive(fp)


def auto_repack_archive(fp:str) -> str:
    """Run a sample with each image format, then automatically pick
    the smallest format to repack the rest of the archive with
    Returns path to repacked archive"""
    results = compare_fmts_archive(fp, quiet=True)
    selection = {"desc":results[1][1], "name":results[1][2]}
    fmt_name = selection['name']
    fmt_desc = selection['desc']
    config.img_format = fmt_name
    return repack_archive(fp)



class Kobo:
    blacklisted_fmts = 'webp webpll'
    prefer_epub = False
    epub_ext = '.kepub.epub'
    epub_properties = (
        (None, 'meta', 'portrait', {'property': 'rendition:spread'}),
        (None, 'meta', 'portrait', {'property': 'rendition:orientation'}),
        (None, 'meta', 'pre-paginated', {'property': 'rendition:layout'}))


class Kindle:
    blacklisted_fmts = 'webp webpll'
    prefer_epub = True
    epub_ext = '.epub'
    epub_properties = (
        (None, 'meta', '', {'name': 'fixed-layout', 'content': 'true'}),
        (None, 'meta', '', {'name': 'book-type', 'content': 'comic'}),
        (None, 'meta', '', {'name': 'zero-gutter', 'content': 'true'}),
        (None, 'meta', '', {'name': 'zero-margin', 'content': 'true'}),
        (None, 'meta', '', {'name': 'ke-border-color', 'content': '#FFFFFF'}),
        (None, 'meta', '', {'name': 'ke-border-width', 'content': '0'}),
        (None, 'meta', '', {'name': 'orientation-lock', 'content': 'portrait'}),
        (None, 'meta', '', {'name': 'region-mag', 'content': 'true'}))


class KoboForma(Kobo):
    nickname = 'KOF'
    desc = 'Kobo Forma/Sage'
    size = (1440, 1920)
    gray = True


class KoboLibra(Kobo):
    nickname = 'KOL'
    desc = 'Kobo Libra 1/2'
    size = (1264, 1680)
    gray = True


class KoboElipsa(Kobo):
    nickname = 'KOE'
    desc = 'Kobo Elipsa/Aura One'
    size = (1404, 1872)
    gray = True


class KoboClaraHD(Kobo): # moi
    nickname = 'KOC'
    desc = 'Kobo Clara HD/2E'
    size = (1072, 1448)
    gray = True


class KoboNia(Kobo):
    nickname = 'KON'
    desc = 'Kobo Nia'
    size = (758, 1024)
    gray = True


class Kindle68(Kindle):
    nickname = 'PW5'
    desc = 'Kindle Paperwhite (11th gen)'
    size = (1236, 1648)
    gray = True


class Kindle300(Kindle):
    nickname = 'PW3'
    desc = 'Kindle Paperwhite (7-10th gen)/Basic (10th gen)'
    size = (1072, 1448)
    gray = True


class Kindle212(Kindle):
    nickname = 'PW2'
    desc = 'Kindle Paperwhite (5-6th gen)'
    size = (758, 1024)
    gray = True


class Kindle167(Kindle):
    nickname = 'KT2'
    desc = 'Kindle Basic (7-8th gen)'
    size = (600, 800)
    gray = True


class KindleOasis(Kindle):
    nickname = 'KOA'
    desc = 'Kindle Oasis'
    size = (1264, 1680)
    gray = True


class KindleVoyage(Kindle):
    nickname = 'KVO'
    desc = 'Kindle Voyage'
    size = (1080, 1440)
    gray = True


ProfileList = (KoboForma, KoboLibra, KoboElipsa, KoboClaraHD, Kindle68,
                 Kindle300, Kindle212, Kindle167, KindleOasis, KindleVoyage)
ProfileDict = {cls.nickname:cls for cls in ProfileList}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import shutil
import tempfile
import time
from functools import partial
from itertools import chain
from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED, ZIP_STORED, BadZipFile

import reCBZ
import reCBZ.config as config
from PIL import Image, UnidentifiedImageError
from reCBZ.formats import *
from reCBZ.util import mylog, map_workers, worker_sigint_CTRL_C, human_sort

# TODO:
# include docstrings

VALID_BOOK_FORMATS:tuple = ('cbz', 'zip', 'epub', 'mobi')
SOURCE_NAME:str = 'Source'
chapter_prefix:str = 'v' # :) :D C:


def write_zip(savepath, chapters):
    new_zip = ZipFile(savepath,'w')
    lead_zeroes = len(str(len(chapters)))
    for i, chapter in enumerate(chapters):
        for page in chapter:

            if len(chapters) > 1: # no parent if there's only one chapter
                dest = Path(f'{chapter_prefix}{i+1:0{lead_zeroes}d}') / page.rel_path
            else:
                dest = Path(page.rel_path)
            mylog(f"ZIP: write '{page.name}' to {dest}")
            if config.compress_zip:
                new_zip.write(page.fp, dest, ZIP_DEFLATED, 9)
            else:
                new_zip.write(page.fp, dest, ZIP_STORED)
    new_zip.comment = str.encode(config.ZIPCOMMENT)
    new_zip.close()
    return savepath


def write_epub(savepath, chapters):
    from reCBZ import epub
    pages = list(chain(*chapters))
    title = Path(savepath).stem
    mylog(f'Write .epub: {title}.epub', progress=True)
    if len(chapters) > 1:
        savepath = epub.multi_chapter_epub(title, chapters)
    else:
        savepath = epub.single_chapter_epub(title, pages)

    # if Config.compress_zip:
    #     ZipFile(savepath, mode='w', compression=ZIP_DEFLATED, compresslevel=9).write(savepath)
    return savepath


def write_mobi(savepath, chapters):
    import subprocess
    try:
        subprocess.run(["kindlegen", "", "/dev/null"], capture_output=True)
    except FileNotFoundError:
        raise OSError("'kindlegen' can't be found. is it installed and in PATH?")
    pass


def get_format_class(name):
    if name in (None, ''): return None
    else:
        try:
            return FormatDict[name]
        except KeyError:
            raise ValueError(f"Invalid format name '{name}'")


@worker_sigint_CTRL_C
def convert_page_worker(source, options, savedir=None):
    start_t = time.perf_counter()
    # page = copy.deepcopy(source)
    page = Page(source.fp) # create a copy

    # ensure file can be opened as image, and that it's a valid format
    try:
        mylog(f'Read file: {page.name}', progress=True)
        log_buff = f'/open:  {page.fp}\n'
        source_fmt = page.fmt
        img = page.img
    except (IOError, UnidentifiedImageError) as err:
        if config.ignore_page_err:
            mylog(f"{page.fp}: can't open file as image, ignoring...'")
            return False, page
        else:
            raise err
    except KeyError as err:
        if config.ignore_page_err:
            mylog(f"{page.fp}: invalid image format, ignoring...'")
            return False, page
        else:
            raise err

    # determine target (new) format
    if options['format']:
        new_fmt = options['format']
    else:
        new_fmt = source_fmt
    page.fmt = new_fmt

    # apply format specific actions
    if new_fmt is Jpeg:
      if not img.mode == 'RGB':
          log_buff += '|trans: mode RGB\n'
          img = img.convert('RGB')

    # transform
    if options['grayscale']:
        log_buff += '|trans: mode L\n' # me lol
        img = img.convert('L')

    if all(options['size']):
        log_buff += f'|trans: resize to {options["size"]}\n'
        width, height = img.size
        new_size = options['size']
        # preserve aspect ratio for landscape images
        if page.landscape:
            new_size = new_size[::-1]
        n_width, n_height = new_size
        # downscaling
        if (width > n_width and height > n_height
            and not options['nodown']):
            img = img.resize((new_size), config.RESAMPLE_TYPE)
        # upscaling
        elif not options['noup']:
            img = img.resize((new_size), config.RESAMPLE_TYPE)

    LossyFmt.quality = options['quality']

    # save
    page.img = img
    ext = page.fmt.ext[0]
    if savedir:
        new_fp = Path.joinpath(savedir, f'{page.stem}{ext}')
    else:
        new_fp = Path.joinpath(page.fp.parents[0], f'{page.stem}{ext}')
    log_buff += f'|trans: {source_fmt.name} -> {new_fmt.name}\n'
    page.save(new_fp)

    end_t = time.perf_counter()
    elapsed = f'{end_t-start_t:.2f}s'
    mylog(f'{log_buff}\\write: {new_fp}: took {elapsed}')
    mylog(f'Save file: {new_fp.name}', progress=True)
    return True, page


class Page():
    def __init__(self, file_name):
        self.fp = Path(file_name)
        # i tried for hours but windows can't correctly pickle the
        # GLOBAL_CACHEDIR, it's not thread safe for whatever reason. some
        # instances will init with a new UUID which can't be compared.
        # this is the least hacky way I could come up with to keep Unix parity
        uuid_part = [part for part in self.fp.parts if reCBZ.CACHE_PREFIX in part]
        global_cache = Path(tempfile.gettempdir()) / uuid_part[0]
        local_cache = global_cache / self.fp.relative_to(global_cache).parts[0]
        self.rel_path = self.fp.relative_to(local_cache)
        self.name = str(self.fp.name)
        self.stem = str(self.fp.stem)
        self._img:Image.Image
        self._fmt = None
        self._closed = True

    @property
    def fmt(self):
        if self._fmt is not None:
            return self._fmt
        else:
            PIL_fmt = self.img.format
            if PIL_fmt is None:
                raise KeyError(f"Image.format returned None")
            elif PIL_fmt == "PNG":
                return Png
            elif PIL_fmt == "JPEG":
                return Jpeg
            elif PIL_fmt == "WEBP":
                # https://github.com/python-pillow/Pillow/discussions/6716
                with open(self.fp, "rb") as file:
                    if file.read(16)[-1:] == b"L":
                        return WebpLossless
                    else:
                        return WebpLossy
            else:
                raise KeyError(f"'{PIL_fmt}': invalid format")

    @fmt.setter
    def fmt(self, new):
        self._fmt = new

    @property
    def img(self):
        if self._closed:
            self._img = Image.open(self.fp)
            self._closed = False
            return self._img
        else:
            return self._img

    @img.setter
    def img(self, new:Image.Image):
        self._img = new
        self._closed = False

    @property
    def size(self):
        return self.img.size

    @property
    def landscape(self):
        if self.size[0] > self.size[1]:
            return True
        else:
            return False

    def save(self, dest):
        self.fmt.save(self.img, dest)
        self.fp = Path(dest)
        self.name = str(self.fp.name)
        self.stem = str(self.fp.stem)
        self._img.close()
        self._closed = True

    def __reduce__(self):
        # pickle pee. pum pa rum
        # https://stackoverflow.com/q/19855156/
        return (self.__class__, (self.fp, ))


class ComicArchive():
    def __init__(self, filename:str):
        mylog('Archive: __init__')
        if Path(filename).exists():
            self.fp:Path = Path(filename)
        else:
            raise ValueError(f"{filename}: invalid path")
        self._page_opt = {}
        self._page_opt['format'] = get_format_class(config.img_format)
        self._page_opt['quality'] = config.img_quality
        self._page_opt['size'] = config.img_size
        self._page_opt['grayscale'] = config.grayscale
        self._page_opt['noup'] = config.no_upscale
        self._page_opt['nodown'] = config.no_downscale
        self._index:list = []
        self._chapter_lengths = []
        self._chapters = []
        self._bad_files = []
        self._cachedir = Path(tempfile.mkdtemp(prefix='book_', dir=reCBZ.GLOBAL_CACHEDIR))

    @property
    def bad_files(self):
        return self._bad_files

    def fetch_pages(self):
        if len(self._index) == 0:
            self._index = list(self.extract())
        return self._index

    def fetch_chapters(self):
        # ensure it's a copy, so we can't delete the original objects
        index_copy = [page for page in self.fetch_pages()]
        if len(self._chapter_lengths) == 0:
            self._chapter_lengths = [len(index_copy)]
        chapters = []
        for length in self._chapter_lengths:
            chapters.append(index_copy[:length])
            del index_copy[:length]
        return chapters

    def extract(self, count:int=0, raw:bool=False) -> tuple:
        try:
            source_zip = ZipFile(self.fp)
        except BadZipFile as err:
            raise ValueError(f"Fatal: '{self.fp}': not a zip file")

        compressed_files = source_zip.namelist()
        assert len(compressed_files) >= 1, 'no files in archive'
        if count > 0:
            # select x images from the middle of the archive, in increments of 2
            if count * 2 > len(compressed_files):
                raise ValueError(f"{self.fp} is smaller than samples * 2")
            delta = int(len(compressed_files) / 2)
            compressed_files = compressed_files[delta-count:delta+count:2]

        mylog(f'Extracting: {self.fp}', progress=True)
        for file in compressed_files:
            source_zip.extract(file, self._cachedir)

        # god bless you Georgy https://stackoverflow.com/a/50927977/
        raw_paths = tuple(filter(Path.is_file, Path(self._cachedir).rglob('*')))
        # solves the need to invert files in EPUB, where the destination can't
        # be inferred from the original filepath. critical, because files are
        # randomly ordered on Windows (probably due to the ZLIB implementation)
        sorted_paths = tuple(human_sort(raw_paths))
        sorted_pages = tuple(Page(path) for path in sorted_paths)

        mylog('', progress=True)
        if raw: return sorted_paths
        else: return sorted_pages

    def add_chapter(self, second_archive, start=None, end=None) -> tuple:
        try:
            assert isinstance(second_archive, ComicArchive)
        except AssertionError:
            raise ValueError('second_archive is not an instance of Archive')

        new_chapter = second_archive.fetch_pages()
        if start:
            try:
                assert type(start) is int
            except AssertionError:
                raise ValueError('start must be an integer')
            # can raise IndexError but we want it to do so
            new_chapter = new_chapter[start:]
        if end:
            try:
                assert type(start) is int
            except AssertionError:
                raise ValueError('end must be an integer')
            new_chapter = new_chapter[:end]

        # ensure chapter1 is populated
        self.fetch_chapters()
        self._chapter_lengths.append(len(new_chapter))
        self._index.extend(new_chapter)
        return tuple(self.fetch_pages())

    def convert_pages(self, fmt=None, quality=None, grayscale=None, size=None) -> tuple:
        # TODO assert values are the right type
        options = dict(self._page_opt)
        if fmt is not None: options['format'] = get_format_class(fmt)
        if quality is not None: options['quality'] = int(quality)
        if grayscale is not None: options['grayscale'] = bool(grayscale)
        if size is not None: options['size'] = size

        worker = partial(convert_page_worker, options=options)
        results = map_workers(worker, self.fetch_pages())

        self._bad_files = [item[1].fp for item in results if item[0] is False]
        self._index = [item[1] for item in results if item[0]]
        mylog('', progress=True)
        return tuple(self._index)

    def compute_fmt_sizes(self) -> tuple:
        def compute_single_fmt(sample_pages, cachedir, fmt) -> tuple:
            fmtdir = Path.joinpath(cachedir, fmt.name)
            Path.mkdir(fmtdir)

            options = dict(self._page_opt) # ensure it's a copy
            options['format'] = fmt
            worker = partial(convert_page_worker, savedir=fmtdir, options=options)
            results = map_workers(worker, sample_pages)

            # pages don't need to be sorted here, as they're discarded
            converted_pages = [item[1] for item in results if item[0]]
            nbytes = sum(page.fp.stat().st_size for page in converted_pages)
            return nbytes, fmt.desc, fmt.name

        # extract images and compute their original size
        # manually call extract so we don't overwrite _pages cache
        source_pages = self.extract(count=config.samples_count)
        nbytes = sum(page.fp.stat().st_size for page in source_pages)
        mylog(f'reference format: {source_pages[0].name}')
        source_fmt = source_pages[0].fmt
        source_fsize = [nbytes, f'{SOURCE_NAME} ({source_fmt.desc})',
                        source_fmt.name]

        # compute the size of each format after converting.
        # one thread per individual format. n processes per thread
        fmt_fsizes = []
        worker = partial(compute_single_fmt, source_pages, self._cachedir)
        results = map_workers(worker, config.allowed_page_formats(), multithread=True)
        fmt_fsizes.extend(results)

        # finally, compare
        # in multidepth lists, sorted compares the first element by default :)
        sorted_fmts = list(sorted(fmt_fsizes))
        sorted_fmts.insert(0, source_fsize)
        mylog(str(sorted_fmts))
        mylog('', progress=True)
        return tuple(sorted_fmts)

    def write_archive(self, book_format='cbz', file_name:str='') -> str:
        if book_format not in VALID_BOOK_FORMATS:
            raise ValueError(f"Invalid format '{book_format}'")

        if file_name != '':
            parent = Path(file_name).parents[0]
            if not (parent.exists() and parent.is_dir()):
                raise ValueError(f"Parent folder '{parent}' does not exist")
            new_path = Path(f'{file_name}.{book_format}')
        else:
            # write to current dir
            new_path = Path(f'{self.fp.stem}.{book_format}')
        if new_path.exists():
            mylog(f'Write .{book_format}: {new_path}', progress=True)
            mylog(f'{new_path} exists, removing...')
            new_path.unlink()

        new_path = str(new_path)
        if book_format == 'cbz':
            return write_zip(new_path, self.fetch_chapters())
        elif book_format == 'zip':
            return write_zip(new_path, self.fetch_chapters())
        elif book_format == 'epub':
            return write_epub(new_path, self.fetch_chapters())
        elif book_format == 'mobi':
            raise NotImplementedError
        else:
            raise ValueError

    def add_page(self, fp, index=-1):
        try:
            assert Path(fp).exists()
            Page(fp)
        except AssertionError:
            raise ValueError(f"can't open {fp}")
        # we want the IndexError
        self._index.insert(index, fp)

    def remove_page(self, index):
        return self._index.pop(index)

    def cleanup(self):
        if self._cachedir.exists():
            mylog(f'cleanup(): {self._cachedir}')
            try:
                shutil.rmtree(self._cachedir)
            except PermissionError:
                mylog(f"PermissionError, couldn't clean {self._cachedir}")



import re
import tempfile
from pathlib import Path
from uuid import uuid4

__version__ = "0.8.0"
CMDNAME = 'recbz'

MODULE_PATH = Path(__file__).resolve().parent

# whether to print the 'recbz.py' title at the beginning
SHOWTITLE = True

# global UUID for files stored in temp, so we can ensure multiple instances
# created under the same process don't delete cache currently used by another
CACHE_PREFIX:str = f'reCBZCACHE_'
GLOBAL_CACHEDIR = Path(tempfile.gettempdir()) / f'{CACHE_PREFIX}{str(uuid4().hex)}'
if not GLOBAL_CACHEDIR.exists():
    GLOBAL_CACHEDIR.mkdir()

IMG_FILES = re.compile('^.*\\.(?!png\\b|webp\\b|jpg\\b|jpeg\\b)\\w*$')
EPUB_FILES = re.compile('^.*(calibre_bookmarks.txt)$|^.*(mimetype)$|.*\\.(?=css\\b|opf\\b|ncx\\b|xhtml\\b|xml\\b)\\w*$')
KEPUB_EPUB = re.compile('^.*(?=\\.kepub\\.epub$)')



import platform
import signal
import textwrap
from functools import wraps
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
from re import split

import reCBZ.config as config


class MPrunnerInterrupt(KeyboardInterrupt):
    """KeyboardInterrupt gracefully caught in MP_runner, please catch me"""


def shorten(*args, width=config.term_width()) -> str:
    text = ' '.join(args)
    return textwrap.shorten(text, width=width, placeholder='...')


def mylog(msg:str, progress=False) -> None:
    if config.loglevel == -1:
        return
    elif config.loglevel > 2:
        print(msg, flush=True)
    elif config.loglevel == 2 and not progress:
        print(msg, flush=True)
    elif config.loglevel == 1 and progress:
        msg = '[*] ' + msg
        msg = shorten(msg)
        print(msg, end='\n', flush=True)
    elif config.loglevel == 0 and progress:
        # no newline (i.e. overwrite line)
        # flush last first
        print('[*]'.ljust(config.term_width()), end='\r')
        msg = '[*] ' + msg
        msg = shorten(msg)
        print(msg, end='\r', flush=True)


def human_sort(lst) -> list:
    """ Sort the given iterable in the way that humans expect."""
    # https://stackoverflow.com/a/2669120/
    if not type(lst[0]) is str:
        lst = [str(i) for i in lst]
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in split('([0-9]+)', key)]
    return sorted(lst, key = alphanum_key)


def human_bytes(b:float) -> str:
    # derived from https://github.com/x4nth055 (MIT)
    suffix = "B"
    FACTOR = 1024
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if b < FACTOR:
            return f"{b:.2f}{unit}{suffix}"
        b /= FACTOR
    return f"{b:.2f}Y{suffix}"


def pct_change(base:float, new:float) -> str:
    diff = new - base
    pct_change = diff / base * 100
    if pct_change >= 0:
        return f"+{pct_change:.2f}%"
    else:
        return f"{pct_change:.2f}%"


def pool_CTRL_C_handler(*args, **kwargs):
    global ctrl_c_entered
    ctrl_c_entered = True


def init_pool():
    # set global variable for each process in the pool:
    global ctrl_c_entered
    global default_sigint_handler
    ctrl_c_entered = False
    default_sigint_handler = signal.signal(signal.SIGINT, pool_CTRL_C_handler)


def worker_sigint_CTRL_C(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not 'ctrl_c_entered' in globals():
            # init_pool hasn't been called because we're not from mp_pool_manager
            # (i.e. single threaded)
            return func(*args, **kwargs)
        global ctrl_c_entered
        if not ctrl_c_entered: # the default
            signal.signal(signal.SIGINT, default_sigint_handler)
            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt:
                ctrl_c_entered = True
                return KeyboardInterrupt()
            finally:
                signal.signal(signal.SIGINT, pool_CTRL_C_handler)
        else:
            return KeyboardInterrupt()
    return wrapper


def map_workers(func, tasks, multithread=False):
    pcount = min(len(tasks), config.pcount())
    if pcount == 1:
        return map(func, tasks)
    elif multithread:
        # mourn the day they inevitably condense the parallel modules in
        # python and I have to recall how any of this works
        with ThreadPool(processes=pcount) as Tpool:
            return Tpool.map(func, tasks)
    else:
        if platform.system == 'windows':
            # this hangs on Unix, but prevents hanging on Windows (insanity)
            # god bless https://stackoverflow.com/a/68695455/
            signal.signal(signal.SIGINT, signal.SIG_IGN)
        with Pool(processes=pcount, initializer=init_pool) as MPpool:
            try:
                return MPpool.map(func, tasks)
            except KeyboardInterrupt:
                mylog("MAY YOUR WOES BE MANY")
                MPpool.terminate()
                mylog("AND YOUR DAYS FEW")
                raise MPrunnerInterrupt()



import os
from importlib import resources

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib
from PIL import Image

from reCBZ.formats import FormatList
from reCBZ.profiles import ProfileDict

# LANCZOS sacrifices performance for optimal upscale quality
RESAMPLE_TYPE = Image.Resampling.LANCZOS
ZIPCOMMENT:str = 'repacked with reCBZ'
_cfg = tomllib.loads(resources.read_text("reCBZ", "defaults.toml"))

overwrite:bool = _cfg["general"]["overwrite"]
ignore_page_err:bool = _cfg["general"]["ignore_page_err"]
force_write:bool = _cfg["general"]["force_write"]
no_write:bool = _cfg["general"]["no_write"]
loglevel:int = _cfg["general"]["loglevel"]
processes:int = _cfg["general"]["processes"]
samples_count:int = _cfg["general"]["samples_count"]
archive_format:str = _cfg["archive"]["archive_format"]
compress_zip:int = _cfg["archive"]["compress_zip"]
right_to_left:bool = _cfg["archive"]["right_to_left"]
img_format:str = _cfg["image"]["img_format"]
img_quality:int = _cfg["image"]["img_quality"]
img_size:tuple = _cfg["image"]["img_size"]
no_upscale:bool = _cfg["image"]["no_upscale"]
no_downscale:bool = _cfg["image"]["no_downscale"]
grayscale:bool = _cfg["image"]["grayscale"]
blacklisted_fmts:str = _cfg["image"]["blacklisted_fmts"]
ebook_profile = None


def pcount() -> int:
    default_value = 4
    if processes > 0:
        return processes
    else:
        logical_cores = os.cpu_count()
        try:
            assert logical_cores is not None
            if logical_cores not in range(1,3):
                logical_cores -= 1 # be kind
            return logical_cores
        except AssertionError:
            return default_value


def term_width() -> int:
    # limit output message width. ignored if verbose
    try:
        TERM_COLUMNS, TERM_LINES = os.get_terminal_size()
        assert TERM_COLUMNS > 0 and TERM_LINES > 0
        if TERM_COLUMNS > 120: max_width = 120
        elif TERM_COLUMNS < 30: max_width = 30
        else: max_width = TERM_COLUMNS - 2
    except (AssertionError, OSError):
        print("[!] Can't determine terminal size, defaulting to 78 cols")
        max_width = 78
    return max_width


def set_profile(name) -> None:
    global blacklisted_fmts, ebook_profile, archive_format, img_size, grayscale
    try:
        profile = ProfileDict[name]
    except KeyError:
        raise ValueError(f"Invalid profile '{name}'")
    grayscale = profile.gray
    img_size = profile.size
    # if profile.prefer_epub:
    archive_format = 'epub'
    ebook_profile = profile
    blacklisted_fmts += profile.blacklisted_fmts


def allowed_page_formats() -> tuple:
    try:
        blacklist = blacklisted_fmts.lower().split(' ')
    except AttributeError: # blacklist is None
        return FormatList
    valid_fmts = tuple(fmt for fmt in FormatList if fmt.name not in blacklist)
    assert len(valid_fmts) >= 1, "valid_formats is 0"
    return valid_fmts


_preload_profile = _cfg["archive"]["ebook_profile"]
if _preload_profile != '':
    set_profile(_preload_profile.upper())
